/******************************************************************************\
 gnofin/exp-latex.c   $Revision: 1.3 $
 Copyright (C) 1999 Darin Fisher

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
\******************************************************************************/

#include "common.h"
#include <string.h>
#include <errno.h>
#include "dialogs.h"
#include "data-if.h"
#include "record-stringizer.h"
#include "latex-export.h"

static const struct
{
  guint        id;
  const gchar *name;
  const gchar  just_ch;
}
fields[] = 
{
  { RECORD_FIELD_DATE,           N_("Date"),        'l' },
  { RECORD_FIELD_TYPE,           N_("Type"),        'l' },
  { RECORD_FIELD_NUMBER,         N_("Number"),      'l' },
  { RECORD_FIELD_LINKED_ACCOUNT, N_("Transfer"),    'l' },
  { RECORD_FIELD_PAYEE,          N_("Payee/payor"), 'l' },
  { RECORD_FIELD_MEMO,           N_("Memo"),        'l' },
  { RECORD_FIELD_CLEARED,        N_("S"),           'l' },
  { RECORD_FIELD_AMOUNT,         N_("Amount"),      'r' },
  { RECORD_FIELD_CLEARED_BAL,    N_("Cleared Bal"), 'r' },
  { RECORD_FIELD_OVERALL_BAL,    N_("Overall Bal"), 'r' },
};

static void
write_text (FILE *file, gchar *text)
{
  gchar reqesc[] = "&#${}_%\"";	/* characters requiring escape prefix '\' in latex 
  			    	 * probably there are other characters that I have
			    	 * forgotten, so these will have to do for now */
  gchar mathch[] = "><";	/* these characters must be set in math mode */
  gchar *c;

  trace ("");

  /* latex filter: necessary to translate, for example, "&" to "\&" */

  for (c=text; *c; ++c)
  {
    if (strchr (reqesc, *c) != NULL)
      fprintf (file, "\\%c", *c);
    else if (strchr (mathch, *c) != NULL)
      fprintf (file, "$%c$", *c);
    else
      fprintf (file, "%c", *c);
  }
}

gboolean
latex_export (GtkWindow *parent, const gchar *filename, const Bankbook *book)
{
  FILE *file;
  const GList *ac, *rc;

  trace ("");

  file = fopen (filename, "wt");
  if (file == NULL)
  {
    dialog_error (parent, _("Unable to create file: %s"), strerror (errno));
    return FALSE;
  }

  /* at some point, it would be cool to surface a lot of the latex
   * style settings.. especially the font style.. and of course the 
   * paper type.. for now i am just going to go with US Letter */

  fprintf (file, "\\documentclass[letterpaper,10pt]{article}\n\n"
  		 "\\setlength{\\textheight}{6in}\n"
		 "\\setlength{\\textwidth}{9in}\n"
		 "\\setlength{\\oddsidemargin}{0in}\n"
		 "\\setlength{\\parindent}{0in}\n"
		 "\\setlength{\\parsep}{0.2in}\n"
		 "\\setlength{\\headheight}{0in}\n"
		 "\\setlength{\\headsep}{0in}\n"
		 "\\setlength{\\topmargin}{0in}\n"
		 "\\setlength{\\topskip}{0in}\n"
		 "\\setlength{\\parskip}{1.2ex}\n\n"
                 "\\begin{document}\n\n"
		 "{\\small\\sf %s}\\hrule\\vspace*{.1in}\n"
		 "{\\small\\sf\n", g_basename (filename));

  for (ac=if_bankbook_get_accounts (book); ac; ac=ac->next)
  {
    Account *account = LIST_DEREF (Account, ac);
    guint i;

    fprintf (file, "{\\bf %s}\n\n" 
    		   "\\begin{tabular}{|",
		   if_account_get_name (account));

    for (i=0; i<sizeof_array (fields); ++i)
      fprintf (file, "%c|", fields[i].just_ch);

    fprintf (file, "}\n\\hline\n");

    for (i=0; i<sizeof_array (fields); ++i)
    {
      fprintf (file, fields[i].name);
      
      if (i == (sizeof_array (fields) - 1))
	fprintf (file, " \\\\\n");
      else
	fprintf (file, " & ");
    }
    fprintf (file, "\\hline\\hline\n");

    for (rc=if_account_get_records (account); rc; rc=rc->next)
    {
      RecordInfo rec = {0};
      Record *record = LIST_DEREF (Record, rc);
      gchar *text;

      if_record_get_info (record, 0, &rec);

      for (i=0; i<sizeof_array (fields); ++i)
      {
        text = stringize_record_field (NULL, 0, fields[i].id, &rec);
	write_text (file, text);

	if (i == (sizeof_array (fields) - 1))
	  fprintf (file, " \\\\\n");
	else
	  fprintf (file, " & ");

	g_free (text);
      }
    }
    fprintf (file, "\\hline\n\\end{tabular}\n");
  }
 
  fprintf (file, "}\\end{document}\n");
  fclose (file);
  return TRUE;
}
