/******************************************************************************\
 gnofin/record.c   $Revision: 1.7 $
 Copyright (C) 1999-2000 Darin Fisher

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
\******************************************************************************/

#include "common.h"
#include <math.h>
#include "record.h"
#include "data-if.h"


/******************************************************************************
 * Record info
 */

guint
record_info_copy (RecordInfo *dest, const RecordInfo *src, guint mask)
{
  trace ("");
  g_return_val_if_fail (src, 0);
  g_return_val_if_fail (dest, 0);
  g_return_val_if_fail (!(mask & RECORD_FIELD_LINKED_RECORD), 0);
  g_return_val_if_fail (!(mask & RECORD_FIELD_OVERALL_BAL), 0);
  g_return_val_if_fail (!(mask & RECORD_FIELD_CLEARED_BAL), 0);

  if (mask == 0)
    mask = RECORD_ALL_WRITABLE_FIELDS;

  if (mask & RECORD_FIELD_TYPE)
    dest->type = src->type;
  
  if (mask & RECORD_FIELD_DATE)
    dest->date = src->date;
  
  if (mask & RECORD_FIELD_NUMBER)
    dest->number = src->number;
  
  if (mask & RECORD_FIELD_LINKED_ACCOUNT)
    dest->linked_account = src->linked_account;

  if (mask & RECORD_FIELD_PAYEE)
    dest->payee = g_strdup (src->payee);

  if (mask & RECORD_FIELD_MEMO)
    dest->memo = g_strdup (src->memo);

  if (mask & RECORD_FIELD_CLEARED)
    dest->cleared = src->cleared;

  if (mask & RECORD_FIELD_AMOUNT)
    dest->amount = src->amount;

  if (mask & RECORD_FIELD_EXCHANGE_RATE)
    dest->exchange_rate = src->exchange_rate;

  return mask;
}

void
record_info_clear (RecordInfo *info, guint mask)
{
  trace ("");
  g_return_if_fail (info);

  if (mask == 0)
    mask = RECORD_ALL_WRITABLE_FIELDS;
  
  if (mask & RECORD_FIELD_PAYEE)
    g_free (info->payee);

  if (mask & RECORD_FIELD_MEMO)
    g_free (info->memo);
}

gboolean
record_info_diff (const RecordInfo *a, const RecordInfo *b, guint mask)
{
  trace ("");
  g_return_val_if_fail (a, TRUE);
  g_return_val_if_fail (b, TRUE);
  g_return_val_if_fail (mask, TRUE);

  if (mask & RECORD_FIELD_TYPE)
    if (a->type != b->type) return TRUE;
  
  if (mask & RECORD_FIELD_DATE)
    if (g_date_compare ((GDate *) &a->date, (GDate *) &b->date)) return TRUE;
  
  if (mask & RECORD_FIELD_NUMBER)
    if (a->number != b->number) return TRUE;

  if (mask & RECORD_FIELD_LINKED_ACCOUNT)
    if (a->linked_account != b->linked_account) return TRUE;
  
  if (mask & RECORD_FIELD_LINKED_RECORD)
    if (a->linked_record != b->linked_record) return TRUE;
  
  if (mask & RECORD_FIELD_PAYEE)
    if (strcmp (a->payee, b->payee)) return TRUE;
  
  if (mask & RECORD_FIELD_MEMO)
    if (strcmp (a->memo, b->memo)) return TRUE;
  
  if (mask & RECORD_FIELD_CLEARED)
    if (a->cleared != b->cleared) return TRUE;
  
  if (mask & RECORD_FIELD_AMOUNT)
    if (a->amount != b->amount) return TRUE;
 
  if (mask & RECORD_FIELD_CLEARED_BAL)
    if (a->cleared_bal != b->cleared_bal) return TRUE;
 
  if (mask & RECORD_FIELD_OVERALL_BAL)
    if (a->overall_bal != b->overall_bal) return TRUE;

  if (mask & RECORD_FIELD_EXCHANGE_RATE)
    if (a->exchange_rate != b->exchange_rate) return TRUE;
  
  return FALSE;
}


/******************************************************************************
 * Record
 */

Record *
record_new ()
{
  Record *record;

  trace ("");

  record = g_new0 (Record, 1);
  record->exchange_rate = 1.0;
  record_ref (record);

  return record;
}

void
record_ref (Record *record)
{
  trace ("%p", record);
  g_return_if_fail (record);

  record->ref_count++;
}

void
record_unref (Record *record)
{
  trace ("%p", record);

  if (record && (--record->ref_count == 0))
    record_destroy (record);
}

void
record_destroy (Record *record)
{
  trace ("%p", record);
  g_return_if_fail (record);

  record->ref_count = 0;

  if (record->parent)
    record_detach (record, TRUE);

  record_type_unref (record->type);
  cached_string_unref (record->payee);
  cached_string_unref (record->memo);

  g_free (record);
}

gboolean
record_attach (Record *record, Account *parent)
{
  GCompareFunc func;

  trace ("%p, %p", record, parent);
  g_return_val_if_fail (record, FALSE);
  g_return_val_if_fail (parent, FALSE);

  func = account_get_record_sort_fcn (parent);
  record->parent = parent;  /* this MUST preceed the insertion b/c
  			     * the compare function depends on being
			     * able to get the parent from the record. */
  parent->records = g_list_insert_sorted (parent->records, record, func);

  /* The parent account now references this record */
  record_ref (record);

  if (parent->parent)
    return record_attach_children (record);
  else 
    return TRUE;
}

gboolean
record_attach_children (Record *record)
{
  Bankbook *book;

  trace ("%p", record);
  g_return_val_if_fail (record, FALSE);
  g_return_val_if_fail (record->parent, FALSE);
  g_return_val_if_fail (record->parent->parent, FALSE);
  g_return_val_if_fail (record->type, FALSE);
  g_return_val_if_fail (record->payee, FALSE);
  g_return_val_if_fail (record->memo, FALSE);

  book = record->parent->parent;

  /* Attach cached strings to the bankbook object */
  record->payee = cached_string_attach (record->payee, &book->payee_cache);
  record->memo = cached_string_attach (record->memo, &book->memo_cache);

  /* Attach the type to the data object */
  record->type = record_type_attach (record->type, book);
  record->type->usage_count++;

  /* If this record has a type that is linked, then we must process
   * the link (ie. generate a new record in the linked account). */
  if (record->type->linked)
  {
    Account *linked_account = record->link.account;
    Record  *linked_record  = record->link.record;

    g_assert (linked_account);
    g_assert (linked_account != record->parent);

    if (linked_record == NULL)
    {
      linked_record = record_copy (record);
      linked_record->link.account = record->parent;
      linked_record->link.record = record;
      record_ref (record);

      record->link.record = linked_record;
    }
    else
    {
      linked_record->date = record->date;
      linked_record->number = record->number;

      /* cleared field is skipped since linked transactions need to be
       * individually cleared in each respective account. 
       *
       * FIXME: we can't make the record fields differ at this time
       *        because the XML module is set up to ignore the first
       *        one it reads!! */

      linked_record->cleared = record->cleared;

      cached_string_unref (linked_record->payee);
      cached_string_unref (linked_record->memo);
      linked_record->payee = cached_string_copy (record->payee);
      linked_record->memo = cached_string_copy (record->memo);
    }

    /* Refresh the linked record fields since they may have changed */
    linked_record->exchange_rate = 1.0 / record->exchange_rate;
    linked_record->amount = -1 * rint ((gfloat) record->amount * record->exchange_rate);

    if (linked_record->parent == NULL)
      record_attach (linked_record, linked_account);
  }

  return TRUE;
}

void
record_detach (Record *record, gboolean full)
{
  trace ("%p %s", record, full ? "full" : "");
  g_return_if_fail (record);
  g_return_if_fail (record->parent);

  /* Finally, we remove this record from its parent's list of records */
  record->parent->records = g_list_remove (record->parent->records, record);
  record->parent = NULL;

  /* The parent account no longer references this record */
  record_unref (record);

  if (full)
    record_detach_children (record);
}

void
record_detach_children (Record *record)
{
  trace ("%p", record);
  g_return_if_fail (record);

  record->type->usage_count--;

  cached_string_detach (record->payee);
  cached_string_detach (record->memo);

  if (record->type->linked)
  {
    Record *linked_record = record->link.record;

    g_assert (linked_record);

    if (linked_record->parent)
      record_detach (linked_record, TRUE);
  }
}

gboolean
record_set_info (Record *record, guint mask, const RecordInfo *info)
{
  /* Very minimal error checking is performed within this routine */

  trace ("");
  g_return_val_if_fail (record, FALSE);
  g_return_val_if_fail (record->parent == NULL, FALSE);
  g_return_val_if_fail (info, FALSE);
  g_return_val_if_fail (!(mask & RECORD_FIELD_LINKED_RECORD), FALSE);
  g_return_val_if_fail (!(mask & RECORD_FIELD_OVERALL_BAL), FALSE);
  g_return_val_if_fail (!(mask & RECORD_FIELD_CLEARED_BAL), FALSE);

  if (mask == 0)
    mask = RECORD_ALL_WRITABLE_FIELDS;

  if (mask & RECORD_FIELD_TYPE)
  {
    if (info->type != record->type)
    {
      record_type_unref (record->type);
      record->type = info->type;
      record_type_ref (record->type);
    }
  }

  if (mask & RECORD_FIELD_DATE)
  {
    g_assert (g_date_valid ((GDate *) &info->date));
    record->date = info->date;
  }

  if (mask & RECORD_FIELD_NUMBER)
    record->number = info->number;

  if (mask & RECORD_FIELD_PAYEE)
  {
    if (record->payee)
    {
      if (record->payee->parent)
      {
	cached_string_unref (record->payee);
	record->payee = cached_string_new (info->payee);
      }
      else
        cached_string_set_string (record->payee, info->payee);
    }
    else
      record->payee = cached_string_new (info->payee);
  }

  if (mask & RECORD_FIELD_MEMO)
  {
    if (record->memo)
    {
      if (record->memo->parent)
      {
	cached_string_unref (record->memo);
	record->memo = cached_string_new (info->memo);
      }
      else
        cached_string_set_string (record->memo, info->memo);
    }
    else
      record->memo = cached_string_new (info->memo);
  }

  if (mask & RECORD_FIELD_CLEARED)
    record->cleared = info->cleared;

  if (mask & RECORD_FIELD_AMOUNT)
    record->amount = info->amount;

  if (mask & RECORD_FIELD_EXCHANGE_RATE)
  {
    if (record->type->linked && (info->exchange_rate <= 0.0))  // not a valid exchange rate!!
    {
      g_warning ("invalid (negative or zero) exchange rate specified!!\n");
      record->exchange_rate = 1.0;
    }
    else
      record->exchange_rate = info->exchange_rate;
  }

  if (mask & RECORD_FIELD_LINKED_ACCOUNT && (info->linked_account != NULL))
  {
    g_assert (record->type->linked);
    record->link.account = info->linked_account;
  }

#if 0  // This now happens inside the attach method
  if (record->link.record)
  {
    /* We have to do the same to the linked record */
    Record *linked_record = record->link.record;

    ((RecordInfo *) info)->amount *= -1;

    linked_record->link.record = NULL;
    record_set_info (linked_record, mask, info);
    linked_record->link.record = record;

    ((RecordInfo *) info)->amount *= -1;
  }
#endif

  return TRUE;
}

void
record_get_info (const Record *record, guint mask, RecordInfo *info)
{
  trace ("");
  g_return_if_fail (record);
  g_return_if_fail (info);

  if (mask == 0)
    mask = (guint) -1;

  if (mask & RECORD_FIELD_TYPE)
    info->type = record->type;
  
  if (mask & RECORD_FIELD_DATE)
    info->date = record->date;
  
  if (mask & RECORD_FIELD_NUMBER)
    info->number = record_number (record);
  
  if (mask & RECORD_FIELD_PAYEE)
    info->payee = record->payee->string;
  
  if (mask & RECORD_FIELD_MEMO)
    info->memo = record->memo->string;

  if (mask & RECORD_FIELD_LINKED_ACCOUNT)
    info->linked_account = record_linked_account (record);

  if (mask & RECORD_FIELD_LINKED_RECORD)
    info->linked_record = record_linked_record (record);

  if (mask & RECORD_FIELD_CLEARED)
    info->cleared = record->cleared ? TRUE : FALSE;
  
  if (mask & RECORD_FIELD_AMOUNT)
    info->amount = record_amount (record);

  if (mask & RECORD_FIELD_OVERALL_BAL)
    info->overall_bal = record->overall_bal;

  if (mask & RECORD_FIELD_CLEARED_BAL)
    info->cleared_bal = record->cleared_bal;

  if (mask & RECORD_FIELD_EXCHANGE_RATE)
    info->exchange_rate = record->exchange_rate;
}

guint
record_number (const Record *record)
{
  trace ("");
  g_return_val_if_fail (record, 0);

  if (record->type && record->type->numbered)
    return record->number;
  else
    return 0;
}

money_t
record_amount (const Record *record)
{
  trace ("");
  g_return_val_if_fail (record, 0);

  switch (record->type->sign)
  {
  case RECORD_TYPE_SIGN_POS:
    return money_abs (record->amount);
  case RECORD_TYPE_SIGN_NEG:
    return -1 * money_abs (record->amount);
  }
  return record->amount;
}

Account *
record_linked_account (const Record *record)
{
  trace ("");
  g_return_val_if_fail (record, NULL);

  if (record->type->linked)
    return record->link.account;
  else
    return NULL;
}

Record *
record_linked_record (const Record *record)
{
  trace ("");
  g_return_val_if_fail (record, NULL);

  if (record->type->linked)
    return record->link.record;
  else
    return NULL;
}

gint
record_index (const Record *record)
{
  g_return_val_if_fail (record, 0);
  g_return_val_if_fail (record->parent, 0);

  return g_list_index (record->parent->records, (gpointer) record);
}

Record *
record_copy (const Record *record)
{
  Record *copy;

  trace ("");
  g_return_val_if_fail (record, NULL);

  copy = record_new ();
  memcpy (copy, record, sizeof (Record));
  copy->ref_count = 1;
  copy->parent = NULL;
  copy->link.record = NULL;
  copy->overall_bal = 0;
  copy->cleared_bal = 0;

  if (copy->type)
    record_type_ref (copy->type);
  if (copy->payee)
    cached_string_ref (copy->payee);
  if (copy->memo)
    cached_string_ref (copy->memo);

  return copy;
}

void
record_dump (const Record *record)
{
  g_print ("    R - %p[rc=%u] {ty=%s,d=%d/%d/%d,n=%u,la=%p,lr=%p,p=%03d,m=%03d,a=%ld,ob=%ld,cb=%ld,c=%u,x=%f}\n",
           record,
	   record->ref_count,
	   record->type ? record->type->name : NULL,
	   record->date.day,
	   record->date.month,
	   record->date.year,
	   record->number,
	   record->link.account,
	   record->link.record,
	   record->payee ? cached_string_index (record->payee) : -1,
	   record->memo ? cached_string_index (record->memo) : -1,
	   record->amount,
	   record->overall_bal,
	   record->cleared_bal,
	   record->cleared,
	   record->exchange_rate);
}

// vim: ts=8 sw=2
