/******************************************************************************\
 gnofin/plugin.c   $Revision: 1.3 $
 Copyright (C) 1999-2000 Darin Fisher

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
\******************************************************************************/

/*
 * Support for dynamically-loaded (plugin) components.
 *
 * Based on code from Gnumeric by:
 *    Tom Dyas (tdyas@romulus.rutgers.edu)
 */

#include "common.h"
#include <unistd.h>
#include <dirent.h>
#include <glib.h>
#include <gmodule.h>
#include <string.h>
#include <libgnome/gnome-util.h>
#include "dialogs.h"
#include "plugin.h"

GList *plugin_list = NULL;

PluginData *
plugin_load (GtkWindow *parent, const gchar *filename)
{
  PluginData *pd;
  const gchar *reason;
  guint res;

  g_assert (filename != NULL);
  
  pd = g_new0 (PluginData, 1);
  if (!pd)
  {
    reason = _("Insufficient Memory");
    goto error;
  }

  pd->handle = g_module_open (filename, 0);

  if (!pd->handle)
  {
    reason = g_module_error ();
    goto error;
  }
  
  if (!g_module_symbol (pd->handle, "init_plugin", (gpointer *) &pd->init_plugin))
  {
    reason = _("Plugin initialization failed [init_plugin() not found]");
    goto error;
  }
  
  if ((res = pd->init_plugin (pd)) != 0)
  {
    switch (res)
    {
    case PLUGIN_ERROR_INVALID_API:
      reason = _("Plugin initialization failed [ERROR_INVALID_API]");
      break;
    default:
      reason = _("Plugin initialization failed [ERROR_UNKNOWN]");
    }
    goto error;
  }

  plugin_list = g_list_append (plugin_list, pd);
  return pd;

error:
  /* cleanup */
  if (pd && pd->handle)
    g_module_close (pd->handle);
  g_free (pd);

  /* report error */
  dialog_error (parent, _("Error loading plugin: %s\n[%s]"), filename, reason);

  return NULL;
}

void
plugin_unload (GtkWindow * parent, PluginData * pd)
{
  g_return_if_fail (pd != NULL);
  
  if (pd->can_unload && !pd->can_unload(pd))
  {
    dialog_error (parent, _("Unable to unload plugin because it is busy"));
    return;
  }
  
  if (pd->free_plugin)
    pd->free_plugin (pd);

  plugin_list = g_list_remove (plugin_list, pd);

  g_module_close (pd->handle);
  g_free (pd);
}

static void
plugin_load_plugins_in_dir (GtkWindow * parent, char * directory)
{
  DIR * d;
  struct dirent * e;
  
  if ((d = opendir (directory)) == NULL)
    return;

  while ((e = readdir (d)) != NULL)
  {
    if (strncmp (e->d_name + strlen (e->d_name) - 3, ".so", 3) == 0)
    {
      char * plugin_name;
      
      plugin_name = g_strconcat (directory, e->d_name, NULL);
      plugin_load (parent, plugin_name);
      g_free (plugin_name);
    }
  }
  closedir (d);
}

static void
load_all_plugins (GtkWindow * parent)
{
  char * plugin_dir;
  char * home_dir = getenv ("HOME");
  
  /* Load the user plugins */
  plugin_dir = g_strconcat (home_dir ? home_dir : "", "/.gnofin/plugins/", NULL);
  plugin_load_plugins_in_dir (parent, plugin_dir);
  g_free (plugin_dir);

  /* Load the system plugins */
  plugin_dir = gnome_unconditional_libdir_file ("gnofin/plugins/");
  plugin_load_plugins_in_dir (parent, plugin_dir);
  g_free (plugin_dir);
}

void
plugins_init (GtkWindow * parent)
{
  if (!g_module_supported())
    return;

  load_all_plugins (parent);
}
